package com.flow.listener;

import com.baomidou.mybatisplus.core.toolkit.StringUtils;
import com.flow.common.core.util.SecurityContextUtil;
import com.flow.common.redis.service.RedisService;
import com.flow.entity.*;
import com.flow.entity.node.Node;
import com.flow.entity.settings.CancelConfig;
import com.flow.entity.settings.Settings;
import com.flow.enums.*;
import com.flow.flowable.cmd.RecordActivityDeleteReasonCmd;
import com.flow.flowable.event.CustomFlowableEngineEvent;
import com.flow.flowable.model.JumpServiceTask;
import com.flow.flowable.utils.ProcessElementUtil;
import com.flow.model.FormInfo;
import com.flow.service.*;
import lombok.extern.slf4j.Slf4j;
import org.flowable.bpmn.model.*;
import org.flowable.common.engine.api.delegate.event.FlowableEngineEntityEvent;
import org.flowable.common.engine.api.delegate.event.FlowableEngineEvent;
import org.flowable.engine.ManagementService;
import org.flowable.engine.RepositoryService;
import org.flowable.engine.RuntimeService;
import org.flowable.engine.TaskService;
import org.flowable.engine.delegate.DelegateExecution;
import org.flowable.engine.delegate.event.AbstractFlowableEngineEventListener;
import org.flowable.engine.delegate.event.FlowableCancelledEvent;
import org.flowable.engine.delegate.event.FlowableProcessStartedEvent;
import org.flowable.engine.delegate.event.impl.FlowableProcessCancelledEventImpl;
import org.flowable.engine.impl.persistence.entity.ExecutionEntityImpl;
import org.flowable.engine.repository.ProcessDefinition;
import org.flowable.engine.runtime.ProcessInstance;
import org.flowable.task.service.impl.persistence.entity.TaskEntityImpl;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;
import org.springframework.transaction.support.TransactionSynchronization;
import org.springframework.transaction.support.TransactionSynchronizationManager;

import java.time.LocalDateTime;
import java.util.Collection;
import java.util.HashMap;
import java.util.Map;
import java.util.Objects;

@Component
@Slf4j
public class GlobalActivityEventListener extends AbstractFlowableEngineEventListener {
    @Autowired
    private FlowDefineService flowDefineService;
    @Autowired
    private FlowTaskService flowTaskService;
    @Autowired
    private TaskService taskService;
    @Autowired
    private ManagementService managementService;
    @Autowired
    private NotifyService notifyService;
    @Autowired
    private FlowInstanceService flowInstanceService;
    @Autowired
    private RedisService<String> redisService;

    @Autowired
    private UserService userService;
    @Autowired
    private RepositoryService repositoryService;

    // 任务创建监听器
    @Override
    protected void taskCreated(FlowableEngineEntityEvent event) {
        TaskEntityImpl entity = (TaskEntityImpl) event.getEntity();
        // 更新任务状态
        FlowTask flowTask = FlowTask.build(entity);
        flowTaskService.save(flowTask);
        DelegateExecution execution = super.getExecution(event);
        FlowElement currentFlowElement = execution.getCurrentFlowElement();
        String nobody = ProcessElementUtil.getExtensionElementValue(currentFlowElement, "nobody");
        String assigneeType = ProcessElementUtil.getExtensionElementValue(currentFlowElement, "assigneeType");
        String same = ProcessElementUtil.getExtensionElementValue(currentFlowElement, "same");

        if (StringUtils.isNotBlank(assigneeType)) {
            if (assigneeType.equals(AssigneeTypeEnum.AUTO_PASS.getType())) {
                // 自动通过
                flowTaskService.autoComplete(entity.getId());
                return;
            } else if (assigneeType.equals(AssigneeTypeEnum.AUTO_REFUSE.getType())) {
                // 自动拒绝
                flowTaskService.autoRefuse(entity.getId());
                return;
            }
        }

        // 审批人为空方式
        if (StringUtils.isBlank(entity.getAssignee()) || "-000000".equals(entity.getAssignee())) {
            entity.setAssignee("admin");
            taskService.setAssignee(entity.getId(), "admin");
            if (StringUtils.isNotBlank(nobody)) {
                if (nobody.equals(ApprovalNobodyEnum.PASS.getNobody())) {
                    flowTaskService.autoComplete(entity.getId(), " 审批人为空");
                } else if (nobody.equals(ApprovalNobodyEnum.REFUSE.getNobody())) {
                    flowTaskService.autoRefuse(entity.getId(), " 审批人为空");
                }
                return;
            }
        }

        // 审批人与发起人是同一人时
        if ("skip".equals(same)) {
            String initiator = execution.getVariable("initiator", String.class);
            if (entity.getAssignee().equals(initiator)) {
                flowTaskService.autoComplete(entity.getId(), " 审批人是发起人");
                return;
            }
        }

        // 审批人重复处理
        String duplicate = execution.getVariable("_DUPLICATE_APPROVAL", String.class);
        if (StringUtils.isNotBlank(duplicate)) {
            String taskDefinitionKey = redisService.hget(event.getProcessInstanceId(), entity.getAssignee(), String.class);
            if (duplicate.equals(DuplicateApprovalEnum.ONCE.getType())) {
                if (StringUtils.isNotBlank(taskDefinitionKey)) {
                    flowTaskService.autoComplete(entity.getId(), " 审批人重复");
                    return;
                }
            } else if (duplicate.equals(DuplicateApprovalEnum.ADJACENT.getType())) {
                if (StringUtils.isNotBlank(taskDefinitionKey)) {
                    UserTask userTask = ProcessElementUtil.findParentElement(currentFlowElement, UserTask.class);
                    if (Objects.nonNull(userTask)) {
                        if (userTask.getId().equals(taskDefinitionKey)) {
                            flowTaskService.autoComplete(entity.getId(), " 审批人重复");
                            return;
                        }
                    }
                }
            }
        }
        // 消息通知
        TransactionSynchronizationManager.registerSynchronization(new TransactionSynchronization() {
            @Override
            public void afterCommit() {
                Notify notify = new Notify();
                // notify.setContent(String.format("您有一条待处理的审批任务【%s】，请及时处理。", entity.getName()));
                String instanceId = entity.getProcessInstanceId();
                FormInfo form = flowInstanceService.getForm(instanceId);
                String fromName = form.getFlowDefine().getName();
                String startUserId = form.getStartUserId();
                User userByUserId = userService.getUserById(Long.valueOf(startUserId));

                notify.setSubject("新的审批任务" + "-" + fromName);

                // 审批节点
                notify.setContent(fromName + "-" + entity.getName());

                // 获取流程定义-判断是否为第一个和最后一个节点
                ProcessDefinition processDefinition = repositoryService.createProcessDefinitionQuery()
                        .processDefinitionId(execution.getProcessDefinitionId())
                        .singleResult();

                // 获取流程定义中的所有流元素-判断是否为第一个和最后一个节点
                BpmnModel bpmnModel = repositoryService.getBpmnModel(processDefinition.getId());
                Collection<FlowElement> flowElements = bpmnModel.getMainProcess().getFlowElements();

                // 获取第一个节点（开始事件）和最后一个节点（结束事件）的 ID-判断是否为第一个和最后一个节点
                String startActivityId = null;
                String endActivityId = null;
                for (FlowElement flowElement : flowElements) {
                    if (flowElement instanceof StartEvent) {
                        startActivityId = flowElement.getId();
                    } else if (flowElement instanceof EndEvent) {
                        endActivityId = flowElement.getId();
                    }
                }

                // 判断是否是第一个节点或最后一个节点
                boolean isFirstNode = currentFlowElement.getId().equals(startActivityId);
                boolean isLastNode = currentFlowElement.getId().equals(endActivityId);


                // // 判断是否是第一个节点或最后一个节点
                // boolean isFirstNode = ProcessElementUtil.isFirstNode(currentFlowElement);
                // boolean isLastNode = ProcessElementUtil.isLastNode(currentFlowElement);
                //
                if (isFirstNode || isLastNode) {
                    // 第一个节点或最后一个节点，消息发送人为当前登录用户
                    notify.setSender(Objects.requireNonNull(SecurityContextUtil.getUserInfo()).getUsername());
                } else {
                    // 其余节点，消息发送人为发起人
                    notify.setSender(userByUserId.getUsername());
                }

                notify.setType(NotifyEnum.TODO);
                // 接收者 entity.getAssignee() 下一节点的审批人
                notify.setReceiver(entity.getAssignee());
                notify.setReceivingTime(LocalDateTime.now());
                notify.setUrl("/flow/todo");
                notifyService.notify(notify);

                log.info("登录用户：{}", SecurityContextUtil.getUserId());
                log.info("Assignee：{}", entity.getAssignee());
                log.info("Owner：{}", entity.getOwner());
                log.info("LocalizedName：{}", entity.getLocalizedName());
                log.info("EventName：{}", entity.getEventName());
            }
        });
    }

    // 任务完成监听器
    @Override
    protected void taskCompleted(FlowableEngineEntityEvent event) {
        log.info("任务完成监听器");
    }


    // 流程创建监听器
    @Override
    protected void processStarted(FlowableProcessStartedEvent event) {
        ExecutionEntityImpl entity = (ExecutionEntityImpl) event.getEntity();
        // 更新流程实例状态
        FlowDefine define = flowDefineService.getDefine(entity.getProcessDefinitionId());
        FlowInstance flowInstance = FlowInstance.buildStatus(entity.getProcessInstanceId(), ProcessStatus.RUNNING);
        Settings settings = define.getSettings();
        CancelConfig cancel = settings.getCancel();
        if (cancel.getEnable()) {
            flowInstance.setCancelDays(cancel.getDays());
        } else {
            flowInstance.setCancelDays(0);
        }
        flowInstanceService.save(flowInstance);
    }

    // 流程完成监听器
    @Override
    protected void processCompleted(FlowableEngineEntityEvent event) {
        ExecutionEntityImpl entity = (ExecutionEntityImpl) event.getEntity();
        String instanceId = entity.getProcessInstanceId();

        // 获取表单标题（实例名称）
        // ProcessInstance processInstance = runtimeService.createProcessInstanceQuery()
        //         .processInstanceId(instanceId)
        //         .singleResult();
        // String formTitle = processInstance.getName();

        // 获取模版名称
        FormInfo form = flowInstanceService.getForm(instanceId);
        String fromName = form.getFlowDefine().getName();

        // 更新流程实例状态
        FlowInstance flowInstance = FlowInstance.buildStatus(entity.getProcessInstanceId(), ProcessStatus.COMPLETED);

        flowInstanceService.updateById(flowInstance);
        redisService.del(event.getProcessInstanceId());

        // 消息通知
        TransactionSynchronizationManager.registerSynchronization(new TransactionSynchronization() {
            @Override
            public void afterCommit() {
                Notify notify = new Notify();
                notify.setSubject("申请已通过");
                // notify.setContent(String.format("您申请的流程《%s》已通过", entity.getName()));
                notify.setSubject("申请已通过" + "-" + fromName);
                notify.setContent(fromName + "-" + "已完结");
                notify.setSender(entity.getStartUserId());
                notify.setType(NotifyEnum.SYSTEM);
                notify.setReceiver(entity.getStartUserId());
                notify.setReceivingTime(LocalDateTime.now());
                notify.setUrl("/flow/submit");
                notifyService.notify(notify);

                log.info("实例开始userId{}", (entity.getStartUserId()));
            }
        });
    }

    // 流程取消监听器
    @Override
    protected void processCancelled(FlowableCancelledEvent event) {
        redisService.del(event.getProcessInstanceId());
        FlowableProcessCancelledEventImpl flowableProcessCancelledEvent = (FlowableProcessCancelledEventImpl) event;
        DelegateExecution execution = flowableProcessCancelledEvent.getExecution();
        Map<String, Object> variables = execution.getVariables();
        System.out.println("variables = " + variables);
    }

    @Override
    protected void custom(FlowableEngineEvent event) {
        CustomFlowableEngineEvent entityEvent = (CustomFlowableEngineEvent) event;
        if (Objects.nonNull(entityEvent)) {
            switch (entityEvent.getCustomFlowableEventType()) {
                case PROCESS_COPY:
                    processCopy(entityEvent);
                    break;
                case ACTIVITY_JUMP:
                    activityJump(entityEvent);
                    break;
                default:
                    break;
            }
        }
    }

    protected void processCopy(CustomFlowableEngineEvent event) {

    }

    protected void activityJump(CustomFlowableEngineEvent event) {
        ExecutionEntityImpl entity = event.getEntity();
        JumpServiceTask jumpServiceTask = (JumpServiceTask) entity.getCurrentFlowElement();
        String targetNode = jumpServiceTask.getTargetNode();
        FlowDefine flowDefine = flowDefineService.getDefine(event.getProcessDefinitionId());
        Node target = flowDefine.findFlowNodeOfType(Node.class, targetNode);
        String deleteReason = String.format("<span>跳转至 <span class='color-primary'>%s</span></span>", target.getName());
        managementService.executeCommand(new RecordActivityDeleteReasonCmd(event.getExecutionId(), entity.getActivityId(), deleteReason));
        redisService.del(event.getProcessInstanceId());
    }


}
